local AddonName, GGL = ...
--[[-----------------------------------------------------------------------------
Label Widget
Displays text and optionally an icon.
-------------------------------------------------------------------------------]]
local Type, Version = "GoldGoblinLootGUIWidget-Timer", 1
local AceGUI = LibStub and LibStub("AceGUI-3.0", true)
if not AceGUI or (AceGUI:GetWidgetVersion(Type) or 0) >= Version then return end

-- Lua APIs
local max, min, mod = math.max, math.min, mod
local select, pairs = select, pairs
local tonumber = tonumber
local abs, floor = math.abs, math.floor
local format = string.format
local C_TimerAfter = C_Timer.After

-- WoW APIs
local STOPWATCH_TIME_UNIT = _G.STOPWATCH_TIME_UNIT

--[[-----------------------------------------------------------------------------
Support functions
-------------------------------------------------------------------------------]]
local SEC_TO_MINUTE_FACTOR = 1/60;
local SEC_TO_HOUR_FACTOR = SEC_TO_MINUTE_FACTOR*SEC_TO_MINUTE_FACTOR;
local TIMER_INTERVAL = 1

local function OnAcquire(self)

end

local function OnRelease(self)
    self:StopTimer(self)
end

local function Timer_SetData(self)
    local timer

    if self.isEndless or self.isCounter then
        timer = self.startTime and (self.timeFunc() - self.startTime) or 0
    else
        timer = self.endTime and (self.endTime - self.timeFunc()) or 0
    end

    --local hour = min(floor(timer*SEC_TO_HOUR_FACTOR), 99)
    --local minute = mod(timer*SEC_TO_MINUTE_FACTOR, 60)
    --local second = mod(timer, 60)
    self.timer_lH:SetFormattedText(STOPWATCH_TIME_UNIT, min(floor(timer*SEC_TO_HOUR_FACTOR), 99))
    self.timer_lM:SetFormattedText(STOPWATCH_TIME_UNIT, mod(timer*SEC_TO_MINUTE_FACTOR, 60))
    self.timer_lS:SetFormattedText(STOPWATCH_TIME_UNIT, mod(timer, 60))
end

--[[-----------------------------------------------------------------------------
Methods
-------------------------------------------------------------------------------]]
local methods = {

    ["OnAcquire"] = function(self)
        self:SetSize()
        self:SetTimeFunction()
	end,

    ["OnRelease"] = function(self)
        self:StopTimer()
    end,

    ["IsCounter"] = function(self, flag)
        self.isCounter = flag
        self.isEndless = false
    end,

    ["StartTimer"] = function(self)
        if self.isEndless then
            if self.isPause then
                -- adjust start time
                if self.timeElapsed then
                    self.startTime = self.timeFunc() - self.timeElapsed
                end
                self.isPause = false
            end
        else
            if not self.duration then return end
            if self.isPause or not self.endTime then
                self.endTime = self.timeFunc() + self.duration
                self.isPause = false
            end
        end
        self.isRunning = true
        if self.isEndless then
            self.startTime = self.timeFunc() - (self.timeElapsed or 0)
        else
            self.startTime = self.timeFunc()
        end
        C_TimerAfter(TIMER_INTERVAL, self.TimerUpdate)
    end,

    ["SetTimeFunction"] = function(self, func)
        self.timeFunc = func or GGL.GetTime
    end,

    ["PauseTimer"] = function(self)
        if self.isEndless then
            self.isRunning = false
            self.isPause = true
            self.timeElapsed = self.timeFunc() - self.startTime
        elseif self.endTime then
            self.isRunning = false
            self.isPause = true
            if self.duration then
                self.duration = self.endTime - self.timeFunc()
            end
        end
        self:Fire("OnTimerPause")
    end,

    ["StopTimer"] = function(self, noTimerReset)
        self.isRunning = false
        self.isPause = false
        self.duration = nil
        self.endTime = nil
        self.startTime = nil
        self.isEndless = false
        if not noTimerReset then
            self:SetDuration(0)
        end
        self:Fire("OnTimerStop")
    end,

    ["FinishTimer"] = function(self, noTimerReset)
        if self.isRunning then
            self:StopTimer(noTimerReset)
            self:Fire("OnTimerFinish")
        end
        if not noTimerReset then
            self:SetDuration(0)
        end
    end,

    ["SetDuration"] = function(self, duration)
        self.duration = duration
        self.isEndless = false
        Timer_SetData(self)
        if self.isRunning then
            if not duration or (duration and duration > 0) then
                self:StartTimer()
            else
                self:StopTimer()
            end
        end
    end,

    ["SetEndless"] = function(self, timeElapsed)
        self.isEndless = true
        self.timeElapsed = timeElapsed or 0
        self.startTime = self.timeFunc() - self.timeElapsed
        Timer_SetData(self)
    end,

    ["SetTime"] = function(self, startTime, endTime)
        if not startTime or not endTime then return end
        self:StopTimer()
        self.endTime = endTime
        self.startTime = startTime
        self.duration = endTime - ( startTime or self.timeFunc() )
        self.isRunning = false
        self.isEndless = false
        Timer_SetData(self)
    end,

    ["SetSize"] = function(self, size)
        local fontName, fontHeight, fontFlags
        local fHeight
        if not size or size == "large" then
            fontName, fontHeight, fontFlags = GameFontHighlightLarge:GetFont()
            fHeight = 20
        elseif size == "small" then
            fontName, fontHeight, fontFlags = GameFontHighlightSmall:GetFont()
            fHeight = 12
        elseif type(size) == "number" then
            fontName, fontHeight, fontFlags = GameFontHighlightSmall:GetFont()
            fontHeight = size
            fHeight = fontHeight + 2
        else
            return self:SetSize()
        end

        local width1, width2 = 0, 0
        for k,v in pairs(self) do
            if type(v) == "table" and v.SetFont then
                v:SetFont(fontName, fontHeight, fontFlags)
                if v.id == 2 then
                    width2 = v:GetUnboundedStringWidth() + 2
                    v:SetWidth(width2)
                elseif v.id == 1 then
                    v:SetText("00")
                    width1 = v:GetUnboundedStringWidth()
                end
            end
        end
        self.frame:SetWidth((3*width1)+(2*width2))
        self.frame:SetHeight(fHeight)
    end,
}

--[[-----------------------------------------------------------------------------
Constructor
-------------------------------------------------------------------------------]]
local function Constructor()
    local widget = {}
    widget.type  = Type

    local frame = CreateFrame("Frame")
    frame:SetWidth(76)
    frame:SetHeight(20)
    widget.frame = frame

    -- minutes
    local timer_lM = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_lM:SetPoint("CENTER",frame,"CENTER",0,0)
    timer_lM:SetWordWrap(false)
    timer_lM:SetJustifyH("CENTER")
    timer_lM:SetJustifyV("MIDDLE")
    timer_lM.id = 1
    widget.timer_lM = timer_lM

    -- minutes ":" RIGHT
    local timer_t2 = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_t2:SetPoint("LEFT",timer_lM,"RIGHT",0,0)
    timer_t2:SetWordWrap(false)
    timer_t2:SetJustifyH("CENTER")
    timer_t2:SetJustifyV("MIDDLE")
    timer_t2:SetText(":")
    timer_t2.id = 2
    widget.timer_t2 = timer_t2

    -- minutes ":" LEFT
    local timer_t1 = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_t1:SetPoint("RIGHT",timer_lM,"LEFT",0,0)
    timer_t1:SetWordWrap(false)
    timer_t1:SetJustifyH("CENTER")
    timer_t1:SetJustifyV("MIDDLE")
    timer_t1:SetText(":")
    timer_t1.id = 2
    widget.timer_t1 = timer_t1

    -- hours
    local timer_lH = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_lH:SetPoint("RIGHT",timer_t1,"LEFT",0,0)
    timer_lH:SetWordWrap(false)
    timer_lH:SetJustifyH("CENTER")
    timer_lH:SetJustifyV("MIDDLE")
    timer_lH.id = 1
    widget.timer_lH = timer_lH

    -- seconds
    local timer_lS = frame:CreateFontString(nil, "BACKGROUND", "GameFontHighlightLarge")
    timer_lS:SetPoint("LEFT",timer_t2,"RIGHT",0,0)
    timer_lS:SetWordWrap(false)
    timer_lS:SetJustifyH("CENTER")
    timer_lS:SetJustifyV("MIDDLE")
    timer_lS.id = 1
    widget.timer_lS = timer_lS

    widget.TimerUpdate = function()
        if not widget.isRunning then return end
        if not widget.isEndless and (widget.endTime <= widget.timeFunc()) then
            widget:FinishTimer()
            widget:StopTimer()
            return
        end
        Timer_SetData(widget)
        C_TimerAfter(TIMER_INTERVAL, widget.TimerUpdate)
    end

    -- methodes
    for method, func in pairs(methods) do
        widget[method] = func
    end

    return AceGUI:RegisterAsWidget(widget)
end

AceGUI:RegisterWidgetType(Type, Constructor, Version)
