-- ###########################
-- EVENTS:
--	"Loot_OnItemLoot" - Loot, itemID, ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag
-- 	"Loot_OnMoneyLoot" - Loot, ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag
-- ###########################

local AddonName, GGL = ...

--local print = function() end

local Loot = {}
GGL.Loot = Loot

local GGL_GetTime = GGL.GetTime

local db, SessionDb

-- ###########################
-- Lua
-- ###########################
local assert, type, select, wipe =
	  assert, type, select, wipe

local str_find, str_match, str_format, str_split =
	  string.find, string.match, string.format, string.split

local bit_band =
	  bit.band


local GetItemInfoInstant, GetLootSlotInfo, GetLootSlotLink, GetLootSlotType, GetLFGCompletionReward, GetLFGCompletionRewardItem =
	  GetItemInfoInstant, GetLootSlotInfo, GetLootSlotLink, GetLootSlotType, GetLFGCompletionReward, GetLFGCompletionRewardItem
local IsWorldQuest = GGL.IS_RETAIL and C_QuestLog.IsWorldQuest or function() return false end

-- ###########################
-- FLAGS
-- ###########################
local MAIN_SOURCE_FLAG = GGL.MAIN_SOURCE_FLAG

local SUB_SOURCE_FLAG = GGL.SUB_SOURCE_FLAG

local CATEGORY_FLAG = GGL.CATEGORY_FLAG

-- ###########################
-- GGL
-- ###########################
local LOOT_SPECIAL_CAST_ITEMS = GGL.LOOT_SPECIAL_CAST_ITEMS

-- ###########################
-- Local
-- ###########################
local EventFrame_OnUpdateSender
local GuidCache = {}
local BagLootIsOpen = false
local PlayerGUID

local LOOT_TIMEOUT = 600
local LOOT_SOURCE_DATA = {
	itemList = {},
}

GGL:AddCallbackHandler(Loot)

-- ########################################
-- Parser
-- ########################################
-- Parse the GUID type
-- [Unit type]-0-[server ID]-[instance ID]-[zone UID]-[ID]-[spawn UID]
local GUID_TYPE_LIST = {
	{ "Creature",       "Creature%-0%-%d+%-%d+%-%d+%-(%d+)%-%x+", 	GGL.MAIN_SOURCE_FLAG.NPC }, -- 0-(%d+)%-(%d+)%-(%d+)%-(%d+)%-(%x+)
	{ "GameObject",     "GameObject%-0%-%d+%-%d+%-%d+%-(%d+)%-%x+",	GGL.MAIN_SOURCE_FLAG.OBJECT },
	{ "Item",           "Item%-%d+%-%d+%-%x+", 						GGL.MAIN_SOURCE_FLAG.ITEM },
}

--Creature-0-1463-2241-22016-162374-00002DA785
--Creature-0-1463-2241-22016-160464-00002EF874
-- mainSourceFlag, sourceID
local function ParseGUID(guid)
	if not guid then return end

	for i = 1, #GUID_TYPE_LIST do
        if str_match(guid, GUID_TYPE_LIST[i][1]) then
			return GUID_TYPE_LIST[i][3], str_match(guid, GUID_TYPE_LIST[i][2])
		end
	end
end

-- Parse the ammount of Copper from the money loot
local PARSE_MONEY_AMOUNT = {
	{COPPER_AMOUNT:gsub("%%d", "(%%d+)"), 1},
	{SILVER_AMOUNT:gsub("%%d", "(%%d+)"), 100},
	{GOLD_AMOUNT:gsub("%%d", "(%%d+)"), 10000},
}
local function ParseMoneyCopper(moneyString)
	if not moneyString then return end
	local coins = 0
	local money,_
	for i = 1, #PARSE_MONEY_AMOUNT do
		_,_,money = str_find(moneyString, PARSE_MONEY_AMOUNT[i][1])
		if tonumber(money or "") then
			coins = coins + (tonumber(money) * PARSE_MONEY_AMOUNT[i][2])
		end
	end
	return coins
end

-- Parse a Tooltip for a specific text
local function ParseToolTipForText(tt, text)
	if not tt or not text then return end
	local tt_name = tt:GetName().."TextLeft"
	local ttText
	for i=2, tt:NumLines() do
		ttText = _G[tt_name..i]:GetText()
		if ttText == text then
			return true
		end
	end
end

local ItemLinkCache = GGL:GetCacheTable()
local ITEM_BONUS_PARSE_STRING = "|Hitem:(%d+):%d*:%d*:%d*:%d*:%d*:%d*:%d*:%d*:%d*:%d*:%d*:(.*)|h%["
local function GetItemIDData(itemLink)
	if not itemLink then return end
	if ItemLinkCache[itemLink] then return ItemLinkCache[itemLink] end
	local itemID, itemBonus = str_match(itemLink, ITEM_BONUS_PARSE_STRING)
	itemID = itemID or GetItemInfoInstant(itemLink)
	itemID = tonumber(itemID)
	if itemBonus then
		local data = { str_split(":", itemBonus) }
		if data then
			-- check if we have any bonus things
			local pass = false
			for i = 1, #data do
				if data[i] and data[i] ~= "" and data[i] ~= "0" then
					pass = true
					break
				end
			end
			-- add bonus
			if pass then
				itemID = itemID..","..itemBonus
			end
		end
	end
	ItemLinkCache[itemLink] = itemID
	return itemID
end

-- ###########################
-- Base
-- ###########################
local FORMAT_PUSH_STRING = {
	item = "%s x %d",
}
local FORMAT_PUSH_FUNCTION = {
	item = function(itemID, qty)
		local _, itemLink = GetItemInfo(itemID)
		return str_format(FORMAT_PUSH_STRING.item, itemLink, qty)
	end,
	money = function(copper)
		return copper
	end
}

local timeOffset = nil;
-- returns a unix timestamp in milliseconds
local function GetTimeNow()
    -- Init
    if not timeOffset then
        timeOffset = floor((GetServerTime() - GetTime()) * 1000);
    end

    return timeOffset + floor(GetTime() * 1000);
end

local function IsValidInterval(value1, value2, delta)
	local bandwidthIn, bandwidthOut, latencyHome = GetNetStats()
    return abs(value1 - value2) <= (latencyHome + delta)
end

local TimeCheck = {
	[MAIN_SOURCE_FLAG.ITEM] = {
		[SUB_SOURCE_FLAG.SCRAPPING] = true,
	}
}

local function PushItemLoot(itemID, ammount, source, itemLog)
	if (LOOT_SOURCE_DATA.itemList and LOOT_SOURCE_DATA.itemList[itemID])
		or (TimeCheck[LOOT_SOURCE_DATA.mainSource] and TimeCheck[LOOT_SOURCE_DATA.mainSource][LOOT_SOURCE_DATA.subSourceFlag] and IsValidInterval(LOOT_SOURCE_DATA.time or 0, GetTimeNow(), 500)) then
			Loot:FireCallbackEvent("Loot_OnItemLoot", Loot, itemID, itemLog, ammount, source, LOOT_SOURCE_DATA.itemID, source.mainSource or LOOT_SOURCE_DATA.mainSource, source.subSourceFlag or LOOT_SOURCE_DATA.itemList.subSourceFlag or LOOT_SOURCE_DATA.subSourceFlag, LOOT_SOURCE_DATA.categoryFlag)
			LOOT_SOURCE_DATA.itemList[itemID] = false
	end
end

local function PushMoneyLoot(ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag)
	if sourceID and mainSourceFlag and subSourceFlag and categoryFlag then
		Loot:FireCallbackEvent("Loot_OnMoneyLoot", Loot, ammount, source, sourceID, mainSourceFlag, subSourceFlag, categoryFlag)
	elseif LOOT_SOURCE_DATA.itemList and LOOT_SOURCE_DATA.itemList[0] then
		Loot:FireCallbackEvent("Loot_OnMoneyLoot", Loot, ammount, LOOT_SOURCE_DATA.itemList[0], LOOT_SOURCE_DATA.itemID, LOOT_SOURCE_DATA.mainSource, LOOT_SOURCE_DATA.itemList.subSourceFlag or LOOT_SOURCE_DATA.subSourceFlag, LOOT_SOURCE_DATA.categoryFlag)
	end
end

-- ###########################
-- LOOT_SOURCE_DATA
-- Handles the clear of the source flags
-- ###########################
local UpdateFrame = CreateFrame("Frame")
local LootSaverOnUpdate

-- Clear all source flags
local function ClearLootSourceSaver(delay)
	if delay then
		UpdateFrame.delay = delay
		UpdateFrame:SetScript("OnUpdate", LootSaverOnUpdate)
	else
		if UpdateFrame.delay then
			UpdateFrame.delay = nil
			UpdateFrame:SetScript("OnUpdate", nil)
		end

		LOOT_SOURCE_DATA.mainSource = nil
		LOOT_SOURCE_DATA.subSourceFlag = nil
		LOOT_SOURCE_DATA.categoryFlag = nil
        LOOT_SOURCE_DATA.itemID = nil
        LOOT_SOURCE_DATA.time = nil
	end
end

function LootSaverOnUpdate(self, elapsed)
    if self.delay then
        self.delay = self.delay - elapsed
        if self.delay <= 0 then
            ClearLootSourceSaver()
        end
	end
end

-- ###########################
-- Events
-- ###########################
local EventFunctions = {}
local EventFrame = CreateFrame("Frame")
local function OnEvent(self, event, ...)
	EventFunctions[event](...)
end
EventFrame:SetScript("OnEvent", OnEvent)

-- list of used Events
local EVENT_LIST = {
	"UNIT_SPELLCAST_START",
	"UNIT_SPELLCAST_SUCCEEDED",
	"LOOT_READY",
	"LOOT_CLOSED",
	"CHAT_MSG_LOOT",
	"CHAT_MSG_MONEY",
	"ISLAND_COMPLETED",
	"GROUP_LEFT",
	"LFG_COMPLETION_REWARD",
	"QUEST_TURNED_IN"
}

if GGL.IS_RETAIL then
	EVENT_LIST[#EVENT_LIST+1] = "SHOW_LOOT_TOAST"
	EVENT_LIST[#EVENT_LIST+1] = "QUEST_LOOT_RECEIVED"
end

for i = 1, #EVENT_LIST do
	EventFrame:RegisterEvent(EVENT_LIST[i])
end

-- mainSourceFlag, sourceID, ammount
local function ProcessLootSources(sources, index, itemID)
	local checkByIndex, _ = type(index) == "number"
	if not itemID then
		if GetLootSlotType(index) == LOOT_SLOT_ITEM  then
			itemID = GetItemInfoInstant(GetLootSlotLink(index))
		elseif GetLootSlotType(index) == LOOT_SLOT_MONEY then
			itemID = 0
		end
	else
		index = index or "saved"
	end

	if itemID then
		local list = LOOT_SOURCE_DATA.itemList
		if not list[itemID] then
			list[itemID] = {
				slots = {}
			}
		end
		list = list[itemID]
		local fullAmount = 0
		if not list.slots[index] then
			local subSourceFlag
			for i = 1, #sources, 2 do
				local mainSourceFlag, sourceID

				if sources[i] and GuidCache[sources[i]] then
					mainSourceFlag, subSourceFlag, sourceID = GuidCache[sources[i]][1], GuidCache[sources[i]][2], GuidCache[sources[i]][4]
					list.mainSource = mainSourceFlag
					list.subSourceFlag = subSourceFlag
				else
					mainSourceFlag, sourceID = ParseGUID(sources[i])
				end
				sourceID = sourceID and tonumber(sourceID) or nil
				-- check if the source is an item / most of the time this means opening
				if mainSourceFlag == MAIN_SOURCE_FLAG.ITEM and (not LOOT_SOURCE_DATA.subSourceFlag or LOOT_SOURCE_DATA.subSourceFlag == MAIN_SOURCE_FLAG.NONE) then
					LOOT_SOURCE_DATA.subSourceFlag = SUB_SOURCE_FLAG.OPENING
					subSourceFlag = SUB_SOURCE_FLAG.OPENING
					list.mainSource = MAIN_SOURCE_FLAG.ITEM
					list.subSourceFlag = subSourceFlag
				end
				-- only check NPC's for multiple sourceID's
				if mainSourceFlag ~= MAIN_SOURCE_FLAG.NPC then
					local lootQuantity
					if checkByIndex then
						_, _, lootQuantity = GetLootSlotInfo(index)
					else
						lootQuantity = sources[i+1] or 1
					end
					list[#list+1] = {mainSourceFlag or LOOT_SOURCE_DATA.mainSource or MAIN_SOURCE_FLAG.NONE, LOOT_SOURCE_DATA.itemID or sourceID, lootQuantity}
					fullAmount = lootQuantity
					list.mainSource = mainSourceFlag or LOOT_SOURCE_DATA.mainSource or MAIN_SOURCE_FLAG.NONE
					list.subSourceFlag = LOOT_SOURCE_DATA.subSourceFlag or subSourceFlag
					break
				else
					list[#list+1] = {mainSourceFlag or LOOT_SOURCE_DATA.mainSource or MAIN_SOURCE_FLAG.NONE, LOOT_SOURCE_DATA.itemID or sourceID, sources[i+1]}
					fullAmount = fullAmount + sources[i+1]
				end
			end

			-- failcheck
			if checkByIndex then
				local lootIcon, lootName, lootQuantity = GetLootSlotInfo(index)
				if lootQuantity and fullAmount and lootQuantity > fullAmount then
					local last = list[#list]
					if last then
						list[#list+1] = {last[1], last[2], lootQuantity-fullAmount}
					else
						list[#list+1] = {LOOT_SOURCE_DATA.mainSource or MAIN_SOURCE_FLAG.NONE, LOOT_SOURCE_DATA.itemID, lootQuantity-fullAmount}
					end
				end
			end

			list.slots[index] = true
		end
	end

end

-- Loot Ready
local lootRedayDelay
function EventFunctions.LOOT_READY()
	if not lootRedayDelay or not IsValidInterval(lootRedayDelay, GetTimeNow(), 300) then
		wipe(LOOT_SOURCE_DATA.itemList)
	end

	for slot=1, GetNumLootItems() do
		-- { sourceGUID1, quant1, sourceGUID2, quant2 }
		local lootSources = { GetLootSourceInfo(slot) }
		local slotType = GetLootSlotType(slot)

		if slotType == LOOT_SLOT_NONE then
			-- nothing, skip
		elseif slotType == LOOT_SLOT_ITEM then
			-- get loot sources
			ProcessLootSources(lootSources, slot)
			-- Fishingloot
			if IsFishingLoot() then
				if LOOT_SOURCE_DATA.mainSource ~= SUB_SOURCE_FLAG.FISHING then
					LOOT_SOURCE_DATA.mainSource 	= MAIN_SOURCE_FLAG.ZONE
					LOOT_SOURCE_DATA.subSourceFlag 	= SUB_SOURCE_FLAG.FISHING
					LOOT_SOURCE_DATA.categoryFlag 	= CATEGORY_FLAG.PROFESSION + CATEGORY_FLAG.FARM_PROFESSION
					LOOT_SOURCE_DATA.itemID 		= nil
				end
			-- last event still valid?
			elseif LOOT_SOURCE_DATA.time and IsValidInterval(LOOT_SOURCE_DATA.time or 0, GetTimeNow(), 1000) then
				-- yes still valid
				----print("Spell")
			elseif LOOT_SOURCE_DATA.mainSource then
				-- normal loot
				----print("Normal loot")
				ClearLootSourceSaver()
			end
		elseif slotType == LOOT_SLOT_MONEY then
			-- nothing, skip
			ProcessLootSources(lootSources, slot)
		elseif slotType == LOOT_SLOT_CURRENCY  then
			-- ignore currency, not needed
		end
	end
	lootRedayDelay = GetTimeNow()
	GGL:LootMob()
end

-- Loot Window Closed
function EventFunctions.LOOT_CLOSED()
	----print("LOOT_CLOSED", GetTimeNow(), LOOT_SOURCE_DATA.time)
	LOOT_SOURCE_DATA.time = nil
    -- clear after a minute
	ClearLootSourceSaver(10)

	BagLootIsOpen = false
end

-- ###########################
-- Set the loot flags
-- Detec some extra things
-- ###########
-- Crafting
-- Gathering
-- PickPocket
-- Chests
-- Scrapping
-- ###########################

-- Scrapping spell ID
local SCRAPPING_SPELL = (C_ScrappingMachineUI and C_ScrappingMachineUI.GetScrapSpellID()) or 265742

-- Tracked Spells
local SPELL_LIST_RAW = {
    -- Professsions
    [13262] = { SUB_SOURCE_FLAG.DISENCHANTING, 	    MAIN_SOURCE_FLAG.ITEM, 	CATEGORY_FLAG.PROFESSION },
    [49383] = { SUB_SOURCE_FLAG.ENGINEERING, 		MAIN_SOURCE_FLAG.NPC, 	CATEGORY_FLAG.PROFESSION },
    [51005] = { SUB_SOURCE_FLAG.MILLING, 			MAIN_SOURCE_FLAG.ITEM, 	CATEGORY_FLAG.PROFESSION },
    [31252] = { SUB_SOURCE_FLAG.PROSPECTING, 		MAIN_SOURCE_FLAG.ITEM, 	CATEGORY_FLAG.PROFESSION },

    -- Farm professions
    [131476] = { SUB_SOURCE_FLAG.FISHING, 	MAIN_SOURCE_FLAG.ZONE, 	    CATEGORY_FLAG.PROFESSION + CATEGORY_FLAG.FARM_PROFESSION },
    [2366] = { SUB_SOURCE_FLAG.HERBALISM, 	MAIN_SOURCE_FLAG.OBJECT, 	CATEGORY_FLAG.PROFESSION + CATEGORY_FLAG.FARM_PROFESSION, "mouseover" },
    [2575] = { SUB_SOURCE_FLAG.MINING, 		MAIN_SOURCE_FLAG.OBJECT, 	CATEGORY_FLAG.PROFESSION + CATEGORY_FLAG.FARM_PROFESSION, "mouseover" },
    [8613] = { SUB_SOURCE_FLAG.SKINNING, 	MAIN_SOURCE_FLAG.NPC, 		CATEGORY_FLAG.PROFESSION + CATEGORY_FLAG.FARM_PROFESSION, "mouseover" },

    -- Classes
    [921] = { SUB_SOURCE_FLAG.PICK_POCKET, 	MAIN_SOURCE_FLAG.NPC, 	    CATEGORY_FLAG.CLASS },

    -- Looting
	[3365] = { SUB_SOURCE_FLAG.OPENING,     MAIN_SOURCE_FLAG.OBJECT,    CATEGORY_FLAG.CHEST },
	[1804] = { SUB_SOURCE_FLAG.OPENING,     MAIN_SOURCE_FLAG.OBJECT,    CATEGORY_FLAG.CHEST }, -- lockpick chest :D
	[280493] = { SUB_SOURCE_FLAG.OPENING,     MAIN_SOURCE_FLAG.OBJECT,    CATEGORY_FLAG.CHEST }, -- Scroll of Unlocking

    -- Scrapping
    [SCRAPPING_SPELL] = { SUB_SOURCE_FLAG.SCRAPPING, 	MAIN_SOURCE_FLAG.ITEM, 	        CATEGORY_FLAG.SCRAPPING },
}

-- Formated spell list for use
-- { searchPattern, lootType, sourceFlag, categoryFlag }
local SPELL_LIST = {}
for k,v in pairs(SPELL_LIST_RAW) do
	SPELL_LIST[#SPELL_LIST+1] = {
		"^"..(GetSpellInfo(k) or k),
		v[1],
		v[2],
		v[3],
		v[4],
	}
end

local SpellNameSaver = {}
-- Check the SPELL_LIST for a spell by name
-- @return  mainSourceFlag, subSourceFlag, categoryFlag
local function CheckForSpell(spellId)
	if SpellNameSaver[spellId] then
		return SpellNameSaver[spellId][3], SpellNameSaver[spellId][2], SpellNameSaver[spellId][4], SpellNameSaver[spellId][5]
	else
		local name = GetSpellInfo(spellId)
		for i = 1, #SPELL_LIST do
			if str_match(name, SPELL_LIST[i][1]) then
				--print("CheckForSpell", SPELL_LIST[i][1])
				SpellNameSaver[spellId] = SPELL_LIST[i]
				return CheckForSpell(spellId)
			end
		end
	end
end

local LastSpellCastStart = {}
function EventFunctions.UNIT_SPELLCAST_START(unit, spellCastGuid, spellId)
	if unit ~= "player" then return end
	--print("UNIT_SPELLCAST_START", unit, spellCastGuid, spellId)
	local _, _, _, unitId = CheckForSpell(spellId)
	if unitId then
		--print("found", GetSpellInfo(spellId), UnitGUID(unitId))
		LastSpellCastStart[spellId] = UnitGUID(unitId)
	elseif LOOT_SPECIAL_CAST_ITEMS[spellId] then
		LastSpellCastStart[spellId] = PlayerGUID
	end
end

local LastSpecialSpell
-- Set flags that are created with UNIT_SPELLCAST_SUCCEEDED from SPELL_LIST
function EventFunctions.UNIT_SPELLCAST_SUCCEEDED(unit, spellCast, spellId)
	if unit ~= "player" then return end
	local mainSourceFlag, subSourceFlag, categoryFlag = CheckForSpell(spellId)

	if mainSourceFlag then
		--print(LastSpellCastStart[spellId])
		----print("UNIT_SPELLCAST_SUCCEEDED", GetTimeNow(), LOOT_SOURCE_DATA.time)
		LOOT_SOURCE_DATA.mainSource = mainSourceFlag
		LOOT_SOURCE_DATA.subSourceFlag = subSourceFlag
        LOOT_SOURCE_DATA.categoryFlag = categoryFlag
		LOOT_SOURCE_DATA.time = GetTimeNow()
		LOOT_SOURCE_DATA.itemID = nil
		if spellId and LastSpellCastStart[spellId] then
			GuidCache[LastSpellCastStart[spellId]] = {mainSourceFlag, subSourceFlag, categoryFlag, nil, GGL_GetTime()}
		end
	elseif LOOT_SPECIAL_CAST_ITEMS[spellId] then
		local src = LOOT_SPECIAL_CAST_ITEMS[spellId].flags
		LOOT_SOURCE_DATA.mainSource = src[2]
		LOOT_SOURCE_DATA.subSourceFlag = src[3]
        LOOT_SOURCE_DATA.categoryFlag = src[4]
		LOOT_SOURCE_DATA.time = GetTimeNow()
		LOOT_SOURCE_DATA.itemID = src[1]
		LastSpecialSpell = LOOT_SPECIAL_CAST_ITEMS[spellId]
		if spellId and LastSpellCastStart[spellId] then
			GuidCache[LastSpellCastStart[spellId]] = {src[2], src[3], src[4], src[1], GGL_GetTime()}
		end
	end
end
-- flags = {GGL.MAIN_SOURCE_FLAG.ITEM, GGL.SUB_SOURCE_FLAG.SPECIAL, GGL.CATEGORY_FLAG.SPECIAL},

-- ###########################
-- Loot Handle
-- ###########################
local LOOT_ITEM_PUSHED_SELF = LOOT_ITEM_PUSHED_SELF:gsub("%%s", "(.+)");
local LOOT_ITEM_PUSHED_SELF_MULTIPLE = LOOT_ITEM_PUSHED_SELF_MULTIPLE:gsub("%%s", "(.+)"):gsub("%%d", "(%%d+)")
local LOOT_ITEM_SELF = LOOT_ITEM_SELF:gsub("%%s", "(.+)")
local LOOT_ITEM_SELF_MULTIPLE = LOOT_ITEM_SELF_MULTIPLE:gsub("%%s", "(.+)"):gsub("%%d", "(%%d+)")

-- the loot!
function EventFunctions.CHAT_MSG_LOOT(msg, arg2, arg3, arg4, msgLootName, arg6, arg7, arg8, arg9, arg10, arg11, unitGUID)
	if unitGUID ~= PlayerGUID then return end
	if ( LOOT_SOURCE_DATA.subSourceFlag == SUB_SOURCE_FLAG.SCRAPPING and LOOT_SOURCE_DATA.time and not IsValidInterval(LOOT_SOURCE_DATA.time or 0, GetTimeNow(), 500) ) then
		--print("No loot source found", msg)
		return
    end

	-- find loot
	local found, _, link, qty = str_find(msg, LOOT_ITEM_SELF_MULTIPLE)
    if not found then
        qty, found, _, link = 1, str_find(msg, LOOT_ITEM_SELF)
    end
	if not found then
        found, _, link, qty = str_find(msg, LOOT_ITEM_PUSHED_SELF_MULTIPLE)
    end
	if not found then
        qty, found, _, link = 1, str_find(msg, LOOT_ITEM_PUSHED_SELF)
	end
	if not link then return GGL:AddIntoErrorLog("CHAT_MSG_LOOT / link", msg) end
    local itemID = GetItemInfoInstant(link)
	qty = tonumber(qty)

	if LastSpecialSpell then
		if LastSpecialSpell[itemID] then
			ProcessLootSources({"loot-special-spell", qty}, "loot-special-spell", itemID)
		end
		LastSpecialSpell = nil
	end

    -- push the item
	if itemID and LOOT_SOURCE_DATA.itemList[itemID] then
		PushItemLoot(itemID, qty, LOOT_SOURCE_DATA.itemList[itemID], GetItemIDData(link))
	end
end

function EventFunctions.CHAT_MSG_MONEY(msg)
	PushMoneyLoot(ParseMoneyCopper(msg))
end

-- ###########################
-- BagItems
-- ###########################
-- some boxes are "different"
function EventFunctions.SHOW_LOOT_TOAST(typeIdentifier, itemLink, quantity, specID, sex, personalLootToast, toastMethod, lessAwesome, upgraded, corrupted)
	if not BagLootIsOpen then return end
	if LOOT_SOURCE_DATA.mainSource ~= MAIN_SOURCE_FLAG.ITEM or LOOT_SOURCE_DATA.subSourceFlag ~= SUB_SOURCE_FLAG.OPENING then return end
	if typeIdentifier == "item" then
		GuidCache["item-toast"] = {MAIN_SOURCE_FLAG.ITEM, SUB_SOURCE_FLAG.OPENING, CATEGORY_FLAG.INVENTORY_CONTAINER, LOOT_SOURCE_DATA.itemID, GGL_GetTime()} -- mainSourceFlag, subSourceFlag, sourceID
		local itemID = GetItemInfoInstant(itemLink)
		ProcessLootSources({"item-toast", quantity}, "toast", itemID)
	elseif typeIdentifier == "money" then

	elseif typeIdentifier == "currency" then

	end
end

-- Search bag for a itemID
local function GetItemBagAndSlot(itemLink)
	local itemID = GetItemInfoInstant(itemLink)
	for bag = 0, NUM_BAG_FRAMES do
		for slot = 1, GetContainerNumSlots(bag) do
			if GetContainerItemID(bag, slot) == itemID then
				return bag, slot, itemID
			end
		end
	end
end

-- Triggers on right click of an item
local function Hook_GetContainerItemLink(itemLink, bag, slot)
	if not itemLink then return end

	if not bag or slot then
		bag, slot = GetItemBagAndSlot(itemLink)
	end

	-- double check
	if bag and slot then
		local _, _, _, _, _, lootable, _, _, _, itemID = GetContainerItemInfo(bag, slot)
		----print(lootable, itemID)
		if lootable then
			GGL_Parse_ToolTip:ClearLines()
			GGL_Parse_ToolTip:SetBagItem(bag, slot)
			-- after lockpick the tooltip update takes some time so check also for the last subSource from casts
			if ParseToolTipForText(GGL_Parse_ToolTip, ITEM_OPENABLE) or LOOT_SOURCE_DATA.subSourceFlag == SUB_SOURCE_FLAG.OPENING then
				ClearLootSourceSaver()
                LOOT_SOURCE_DATA.itemID = itemID
                LOOT_SOURCE_DATA.mainSource = MAIN_SOURCE_FLAG.ITEM
                LOOT_SOURCE_DATA.subSourceFlag = SUB_SOURCE_FLAG.OPENING
				LOOT_SOURCE_DATA.categoryFlag = CATEGORY_FLAG.INVENTORY_CONTAINER
				LOOT_SOURCE_DATA.time = GetTimeNow()
				BagLootIsOpen = true
			end
		end
	end
end

-- ###########################
-- Islands
-- ###########################
local IsIslandReward = false
function EventFunctions.ISLAND_COMPLETED()
	IsIslandReward = true
end

function EventFunctions.GROUP_LEFT()
	IsIslandReward = false
end

function EventFunctions.LFG_COMPLETION_REWARD()
	if not IsIslandReward then return end
	local name, typeID, subtypeID, iconTextureFile, moneyBase, moneyVar, experienceBase, experienceVar, numStrangers, numRewards = GetLFGCompletionReward()
	if not numRewards or numRewards == 0 then
		return
	end

	ClearLootSourceSaver()
	LOOT_SOURCE_DATA.mainSource = MAIN_SOURCE_FLAG.ZONE
	LOOT_SOURCE_DATA.subSourceFlag = SUB_SOURCE_FLAG.ISLAND_EXPEDITION
	LOOT_SOURCE_DATA.categoryFlag = CATEGORY_FLAG.LOOT
	LOOT_SOURCE_DATA.time = GetTimeNow()

	for i = 1, numRewards do
		local texture, quantity, isBonus, bonusQuantity, name, quality, itemID, objectType = GetLFGCompletionRewardItem(i);
		if objectType == "item" then
			GuidCache["item-lfg-complete"] = GuidCache["item-lfg-complete"] or {MAIN_SOURCE_FLAG.ZONE, SUB_SOURCE_FLAG.ISLAND_EXPEDITION, CATEGORY_FLAG.LOOT, nil, GGL_GetTime()} -- mainSourceFlag, subSourceFlag, sourceID
			ProcessLootSources({"item-lfg-complete", quantity}, "item-lfg-complete", itemID)
		end
	end
end

-- ###########################
-- Quests
-- ###########################
function EventFunctions.QUEST_LOOT_RECEIVED(questID, itemLink, quantity)
	local subSource = IsWorldQuest(questID) and SUB_SOURCE_FLAG.WORLD_QUEST or SUB_SOURCE_FLAG.QUEST

	GuidCache["quest-loot"] = {MAIN_SOURCE_FLAG.ZONE, subSource, CATEGORY_FLAG.QUEST, questID, GGL_GetTime()} -- mainSourceFlag, subSourceFlag, sourceID
	local itemID = GetItemInfoInstant(itemLink)
	ProcessLootSources({"quest-loot", quantity}, "quest", itemID)
end

function EventFunctions.QUEST_TURNED_IN(questID, xpReward, moneyReward)
	local subSource = IsWorldQuest(questID) and SUB_SOURCE_FLAG.WORLD_QUEST or SUB_SOURCE_FLAG.QUEST

	PushMoneyLoot(moneyReward, nil, questID, MAIN_SOURCE_FLAG.ZONE, subSource, CATEGORY_FLAG.QUEST)
end

-- ###########################
-- Cache clearer
-- ###########################
-- GGL_GetTime
local CacheClearTimer = GGL:GetTimer()
local CACHE_CLEAR_TIME = 10*60
CacheClearTimer:ManagedAfter(60, function()
	local timeNow = GGL_GetTime()
	for k,v in pairs(GuidCache) do
		if not v[5] or ((timeNow - v[5]) > CACHE_CLEAR_TIME) then
			GuidCache[k] = nil
		end
	end
	return true
end)

-- ###########################
-- Init
-- ###########################
local function Init()

	hooksecurefunc("UseItemByName", function(name, target)
        if not target then
            Hook_GetContainerItemLink(select(2, GetItemInfo(name)))
        end
    end)

    hooksecurefunc("UseContainerItem", function(bag, slot, target)
        if not target then
            Hook_GetContainerItemLink(GetContainerItemLink(bag, slot), bag, slot)
        end
	end)

	PlayerGUID = UnitGUID("player")
end
GGL:AddInitFuncPlayerLogin(Init)
